using System;
using System.Web;
using System.Web.Configuration;
using System.Configuration;
using System.Configuration.Provider;
using System.Collections;
using System.Collections.Specialized;
using System.Web.SessionState;
using System.Diagnostics;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;




namespace SoftDweller.Session
{

    public sealed class FileSessionStateStore : SessionStateStoreProviderBase
    {
        private SessionStateSection pConfig = null;
        private string pApplicationName;
        private string pFilePath = string.Empty;
        private SessionStateItemExpireCallback m_endcallback = null;
        private string pFileExt = ".session";
        private string pStaticFileExt = ".static";
        private System.Threading.Timer clearfilestimer = null;
        private TimeSpan pFilesExpire = TimeSpan.Parse("1.00:00:00");
        private System.Security.Principal.WindowsIdentity pWUser = System.Security.Principal.WindowsIdentity.GetCurrent();

        //
        // The ApplicationName property is used to differentiate sessions
        // in the data source by application.
        //

        public string ApplicationName
        {
            get { return pApplicationName; }
        }

        public string SessionFilePath
        {
            get { return pFilePath; }
        }

        //
        //Check The Session File Folder. Create the Folder
        //
        //

        private void CheckFilePath()
        {
            if (Directory.Exists(pFilePath.TrimEnd("\\".ToCharArray())) == false)
                Directory.CreateDirectory(pFilePath.TrimEnd("\\".ToCharArray()));
        }


        public override void Initialize(string name, NameValueCollection config)
        {
            //
            // Initialize values from web.config.
            //

            pWUser = System.Security.Principal.WindowsIdentity.GetCurrent();

            System.Threading.Thread.CurrentPrincipal = new System.Security.Principal.WindowsPrincipal(pWUser);

            if (config == null)
                throw new ArgumentNullException("config");

            if (name == null || name.Length == 0)
                name = "FileSessionStateStore";

            if (string.IsNullOrEmpty(config["description"]))
            {
                config.Remove("description");
                config.Add("description", "File System with cache Session State Store provider");
            }

            // Initialize the abstract base class.
            base.Initialize(name, config);


            //
            // Initialize the ApplicationName property.
            //

            pApplicationName =
                System.Web.Hosting.HostingEnvironment.ApplicationVirtualPath;


            //
            // Get <sessionState> configuration element.
            //

            Configuration cfg =
                WebConfigurationManager.OpenWebConfiguration(ApplicationName);
            pConfig =
                (SessionStateSection)cfg.GetSection("system.web/sessionState");


            //
            // Initialize session files directory.
            //
            if (string.IsNullOrEmpty(config["FilePath"]))
                pFilePath = System.Web.Hosting.HostingEnvironment.ApplicationPhysicalPath + @"App_Data\_aspnet_session_files\";
            else
            {
                pFilePath = config["FilePath"];
                if (pFilePath.EndsWith(@"\") == false)
                    pFilePath += @"\";
            }

            CheckFilePath();

            //
            // Set the clear session files timer
            //

            DateTime cftime = DateTime.MinValue;
            if (string.IsNullOrEmpty(config["ClearFilesTime"]))
                cftime = DateTime.Parse("05:30:00");
            else
                cftime = DateTime.Parse(config["ClearFilesTime"]);

            TimeSpan tfirstrun = cftime.Subtract(DateTime.Now);

            if (tfirstrun.TotalSeconds <= 0)
                tfirstrun = tfirstrun.Add(TimeSpan.FromDays(1));

            if (string.IsNullOrEmpty(config["FilesExpire"]) == false)
                pFilesExpire = TimeSpan.Parse(config["FilesExpire"].ToString());

            string cfstr = "1.00:00:00";
            if (string.IsNullOrEmpty(config["ClearFilesInterval"]))
                cfstr = "1.00:00:00";
            else
                cfstr = config["ClearFilesInterval"].ToString();

            bool enablesche = true;
            if (string.IsNullOrEmpty(config["EnableClearFilesScheduler"]) == false)
                enablesche = bool.Parse(config["EnableClearFilesScheduler"]);

            if(enablesche)
                clearfilestimer = new System.Threading.Timer(new System.Threading.TimerCallback(ClearFilesAction), pFilePath + "\n" + pFilesExpire.ToString(), tfirstrun, TimeSpan.Parse(cfstr));

            bool cfwss = false;
            if (string.IsNullOrEmpty(config["ClearFilesWhenServerStarted"]) == false)
                cfwss = bool.Parse(config["ClearFilesWhenServerStarted"]);

            if (cfwss)
            {
                System.Threading.Thread snthread = new System.Threading.Thread(new System.Threading.ParameterizedThreadStart(ClearFilesAction));
                snthread.Start(pFilePath + "\n" + pFilesExpire.ToString());
            }

            
        }


        //
        // SessionStateStoreProviderBase members
        //

        public override void Dispose()
        {
        }


        //
        // SessionStateProviderBase.SetItemExpireCallback
        //

        public override bool SetItemExpireCallback(SessionStateItemExpireCallback expireCallback)
        {
            this.m_endcallback = expireCallback;
            return true;
        }

        private string GetSessionCacheId(string id)
        {
            return "__ASP.NET.SESSION__" + ApplicationName + "__SESSIONID__" + id + "__UNIQUE__{7D02BC1B-977F-47c7-88DF-973259756F98}";
        }


        //
        // SessionStateProviderBase.SetAndReleaseItemExclusive
        //

        public override void SetAndReleaseItemExclusive(HttpContext context,
            string id,
            SessionStateStoreData item,
            object lockId,
            bool newItem)
        {
            
            SessionPackage sessionp = null;

            if (context.Cache.Get(GetSessionCacheId(id)) == null)
                sessionp = new SessionPackage(id, this.ApplicationName, (SessionStateItemCollection)item.Items);
            else
                sessionp = (SessionPackage)context.Cache.Get(GetSessionCacheId(id));

            sessionp.SessionId = id;
            sessionp.ApplicationName = this.ApplicationName;
            sessionp.SessionItems = (SessionStateItemCollection)item.Items;
            if (newItem)
            {
                sessionp.Created = DateTime.Now;
                sessionp.LockDate = DateTime.Now;
                sessionp.LockId = Guid.NewGuid().ToString();
            }
            sessionp.Expires = DateTime.Now.AddMinutes(Convert.ToDouble(item.Timeout));
            sessionp.Timeout = item.Timeout;
            sessionp.Locked = false;
            sessionp.Flags = 0;


            SessionCacheRemoveAction scra = new SessionCacheRemoveAction(m_endcallback, id, new SessionStateStoreData(item.Items, sessionp.StaticObjects, item.Timeout));

            context.Cache.Insert(GetSessionCacheId(id),
                sessionp,
                null,
                DateTime.MaxValue,
                TimeSpan.FromMinutes(Convert.ToDouble(item.Timeout)),
                System.Web.Caching.CacheItemPriority.Low,
                new System.Web.Caching.CacheItemRemovedCallback(scra.DoCacheRemoveEvent));

            System.Threading.Thread savet = new System.Threading.Thread(new System.Threading.ParameterizedThreadStart(this.ThreadedSave));
            savet.Start(sessionp);



        }


        //
        // SessionStateProviderBase.GetItem
        //

        public override SessionStateStoreData GetItem(HttpContext context,
            string id,
            out bool locked,
            out TimeSpan lockAge,
            out object lockId,
            out SessionStateActions actionFlags)
        {
            return GetSessionStoreItem(false, context, id, out locked,
                out lockAge, out lockId, out actionFlags);
        }


        //
        // SessionStateProviderBase.GetItemExclusive
        //

        public override SessionStateStoreData GetItemExclusive(HttpContext context,
            string id,
            out bool locked,
            out TimeSpan lockAge,
            out object lockId,
            out SessionStateActions actionFlags)
        {
            return GetSessionStoreItem(true, context, id, out locked,
                out lockAge, out lockId, out actionFlags);
        }


        //
        // GetSessionStoreItem is called by both the GetItem and 
        // GetItemExclusive methods. GetSessionStoreItem retrieves the 
        // session data from the data source. If the lockRecord parameter
        // is true (in the case of GetItemExclusive), then GetSessionStoreItem
        // locks the record and sets a new LockId and LockDate.
        //

        private SessionStateStoreData GetSessionStoreItem(bool lockRecord,
            HttpContext context,
            string id,
            out bool locked,
            out TimeSpan lockAge,
            out object lockId,
            out SessionStateActions actionFlags)
        {
            // Initial values for return value and out parameters.
            SessionStateStoreData item = null;
            lockAge = TimeSpan.Zero;
            lockId = null;
            locked = false;
            actionFlags = 0;

            // DateTime to check if current session item is expired.
            DateTime expires;
            // String to hold serialized SessionStateItemCollection.
            SessionStateItemCollection cachedItems = new SessionStateItemCollection();
            
            // Timeout value from the data store.
            int timeout = 0;


            bool fileisnew, fileislocked;

            SessionPackage sessionp = (SessionPackage)context.Cache.Get(GetSessionCacheId(id));

            if (sessionp == null)
            {

                sessionp = MountSession(context, id, out fileisnew, out fileislocked);

                if (fileisnew)
                    actionFlags = SessionStateActions.InitializeItem;
            }

            // lockRecord is true when called from GetItemExclusive and
            // false when called from GetItem.
            // Obtain a lock if possible. Ignore the record if it is expired.

            if (lockRecord)
            {
                if (sessionp.Locked)
                {
                    locked = true;
                    lockId = sessionp.LockId;
                    lockAge = DateTime.Now.Subtract(sessionp.LockDate);

                    return null;
                }
                else
                {

                    locked = false;

                    sessionp.Locked = true;
                    sessionp.LockDate = DateTime.Now;
                }
            }
            // No record was updated because the record was locked or not found.        

            // The record was updated.






            // Retrieve the current session item information.
            if (sessionp != null)
            {
                expires = sessionp.Expires;

                cachedItems = sessionp.SessionItems;
                lockId = sessionp.LockId;
                lockAge = DateTime.Now.Subtract(sessionp.LockDate);
                actionFlags = (SessionStateActions)sessionp.Flags;
                timeout = sessionp.Timeout;
            }
           

           

            // If the record was found and you obtained a lock, then set 
            // the lockId, clear the actionFlags,
            // and create the SessionStateStoreItem to return.
            if (!locked)
            {
                // lockId = Convert.ToInt32(lockId) + 1;
                sessionp.LockId = Guid.NewGuid().ToString();

                
                sessionp.Flags = 0;

                // If the actionFlags parameter is not InitializeItem, 
                // deserialize the stored SessionStateItemCollection.
                
                item = new SessionStateStoreData(sessionp.SessionItems, SessionStateUtility.GetSessionStaticObjects(context), sessionp.Timeout);
            }
            //}

            SessionCacheRemoveAction scran = new SessionCacheRemoveAction(m_endcallback, id, new SessionStateStoreData(sessionp.SessionItems, sessionp.StaticObjects, sessionp.Timeout));

            context.Cache.Insert(GetSessionCacheId(id),
                sessionp,
                null,
                DateTime.MaxValue,
                TimeSpan.FromMinutes(Convert.ToDouble(sessionp.Timeout)),
                System.Web.Caching.CacheItemPriority.Low,
                new System.Web.Caching.CacheItemRemovedCallback(scran.DoCacheRemoveEvent));


            return item;
        }

        //
        //Mount Back The Whole Session into the cache from file
        //

        private SessionPackage MountSession(HttpContext context, string id, out bool isnew, out bool islocked)
        {
            SessionPackage mountedsession = new SessionPackage();
            mountedsession.SessionId = id;
            mountedsession.ApplicationName = ApplicationName;
            mountedsession.Created = DateTime.Now;
            mountedsession.Expires = DateTime.Now.AddMinutes(Convert.ToDouble(pConfig.Timeout.Minutes));
            mountedsession.LockDate = DateTime.Now;
            mountedsession.LockId = Guid.NewGuid().ToString();
            mountedsession.Timeout = pConfig.Timeout.Minutes;

            mountedsession.SessionItems = MountSessionItemPack(context, id, out isnew, out islocked);
            mountedsession.StaticObjects = MountStaticObjectPack(context, id);

            

            mountedsession.Flags = isnew ? 1 : 0;
            mountedsession.Locked = false;


            return mountedsession;
        }

        //
        //Mount the session data back to the cache
        //
        private SessionStateItemCollection MountSessionItemPack(HttpContext context, string sessionid, out bool isnew, out bool islocked)
        {
            string sess_file = pFilePath + sessionid + pFileExt;
            if (File.Exists(sess_file))
            {
                FileInfo fil = new FileInfo(sess_file);

                if (fil.Length > 0)
                {
                    FileStream fs = null;
                    //FileStream fs = new FileStream( sess_file , FileMode.OpenOrCreate , FileAccess.ReadWrite , FileShare.None , (fil.Length > int.MaxValue) ?  int.MaxValue:Convert.ToInt32(fil.Length) , FileOptions.RandomAccess);
                    while (true)
                    {
                        try
                        {
                            fs = File.Open(sess_file, FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.None);
                        }
                        catch (IOException)
                        {
                            System.Threading.Thread.Sleep(1);
                            continue;
                        }
                        break;
                    }

                    SessionStateItemCollection sssi = Deserialize(context, fs);

                    fs.Close();

                    isnew = false;
                    islocked = false;
                    return sssi;
                }
            }
            isnew = true;
            islocked = false;
            return new SessionStateItemCollection();

        }

        private HttpStaticObjectsCollection MountStaticObjectPack(HttpContext context, string sessionid)
        {
            string sess_file = pFilePath + sessionid + pStaticFileExt;
            if (File.Exists(sess_file))
            {
                FileInfo fil = new FileInfo(sess_file);

                if (fil.Length > 0)
                {
                    FileStream fs = null;
                    //FileStream fs = new FileStream( sess_file , FileMode.OpenOrCreate , FileAccess.ReadWrite , FileShare.None , (fil.Length > int.MaxValue) ?  int.MaxValue:Convert.ToInt32(fil.Length) , FileOptions.RandomAccess);
                    while (true)
                    {
                        try
                        {
                            fs = File.Open(sess_file, FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.None);
                        }
                        catch (IOException)
                        {
                            System.Threading.Thread.Sleep(1);
                            continue;
                        }
                        break;
                    }

                    HttpStaticObjectsCollection sssi = DeserializeObj(context, fs);

                    fs.Close();

                    return sssi;
                }
            }

            return new HttpStaticObjectsCollection();

        }

        //
        //Save the session data into file from the cache
        //
        private bool SaveSessionPack(SessionStateItemCollection items, string sessionid, HttpStaticObjectsCollection staticobjs)
        {
            //Save the Session Items File
            string sess_file = pFilePath + sessionid + pFileExt;
            FileStream fs = null;

            while (true)
            {
                try
                {
                    fs = File.Open(sess_file, FileMode.Create, FileAccess.ReadWrite, FileShare.None);
                }
                catch (IOException)
                {
                    System.Threading.Thread.Sleep(1);
                    continue;
                }

                break;
            }

            bool sresult = Serialize(items, fs);

            fs.Close();

            //Save the Staic Objects File

            string obj_file = pFilePath + sessionid + pStaticFileExt;
            bool oresult = true;

            if (staticobjs.Count > 0 || File.Exists(obj_file))
            {
                while (true)
                {
                    try
                    {
                        fs = File.Open(obj_file, FileMode.Create, FileAccess.ReadWrite, FileShare.None);
                    }
                    catch (IOException)
                    {
                        System.Threading.Thread.Sleep(1);
                        continue;
                    }

                    break;
                }

                oresult = Serialize(staticobjs, fs);

                fs.Close();
            }

            return (sresult && oresult);
        }

        //
        // Serialize is called by the SetAndReleaseItemExclusive method to 
        // convert the SessionStateItemCollection into a Base64 string to    
        // be stored in an Access Memo field.
        //

        private bool Serialize(SessionStateItemCollection items, FileStream ms)
        {
            BinaryWriter writer = new BinaryWriter(ms);

            if (items != null)
            {
                items.Serialize(writer);
            }

            writer.Close();

            return true;
        }

        private bool Serialize(HttpStaticObjectsCollection items, FileStream ms)
        {
            BinaryWriter writer = new BinaryWriter(ms);

            if (items != null)
            {
                items.Serialize(writer);
            }

            writer.Close();

            return true;
        }

        //
        // DeSerialize is called by the GetSessionStoreItem method to 
        // convert the Base64 string stored in the Access Memo field to a 
        // SessionStateItemCollection.
        //

        private SessionStateItemCollection Deserialize(HttpContext context, FileStream ms)
        {

            SessionStateItemCollection sessionItems = new SessionStateItemCollection();

            if (ms.Length > 0)
            {
                BinaryReader reader = new BinaryReader(ms);

                try
                {
                    sessionItems = SessionStateItemCollection.Deserialize(reader);
                }
                catch (Exception)
                {
                    reader.Close();
                    return new SessionStateItemCollection();
                }

                reader.Close();

            }

            return sessionItems;
        }

        private HttpStaticObjectsCollection DeserializeObj(HttpContext context, FileStream ms)
        {

            HttpStaticObjectsCollection sessionItems = new HttpStaticObjectsCollection();

            if (ms.Length > 0)
            {
                BinaryReader reader = new BinaryReader(ms);

                try
                {
                    sessionItems = HttpStaticObjectsCollection.Deserialize(reader);
                }
                catch (Exception)
                {
                    reader.Close();
                    return new HttpStaticObjectsCollection();
                }

                reader.Close();

            }

            return sessionItems;
        }

        //
        // SessionStateProviderBase.ReleaseItemExclusive
        //

        public override void ReleaseItemExclusive(HttpContext context,
            string id,
            object lockId)
        {
            //CheckFilePath();

            SessionPackage sessionp = (SessionPackage)context.Cache.Get(GetSessionCacheId(id));

            if (sessionp != null)
            {

                sessionp.Locked = false;

                SessionCacheRemoveAction scra = new SessionCacheRemoveAction(m_endcallback, id, new SessionStateStoreData(sessionp.SessionItems, sessionp.StaticObjects, sessionp.Timeout));

                context.Cache.Insert(GetSessionCacheId(id),
                    sessionp,
                    null,
                    DateTime.MaxValue,
                    TimeSpan.FromMinutes(Convert.ToDouble(sessionp.Timeout)),
                    System.Web.Caching.CacheItemPriority.Low,
                    new System.Web.Caching.CacheItemRemovedCallback(scra.DoCacheRemoveEvent));
            }


        }

        //
        // SessionStateProviderBase.RemoveItem
        //

        public override void RemoveItem(HttpContext context,
            string id,
            object lockId,
            SessionStateStoreData item)
        {

            context.Cache.Remove(GetSessionCacheId(id));

            string sess_file = pFilePath + id + pFileExt;

            if (File.Exists(sess_file))
                File.Delete(sess_file);

            string obj_file = pFilePath + id + pStaticFileExt;

            if (File.Exists(obj_file))
                File.Delete(obj_file);

        }



        //
        // SessionStateProviderBase.CreateUninitializedItem
        //

        public override void CreateUninitializedItem(HttpContext context,
            string id,
            int timeout)
        {
            

            if (context.Cache.Get(GetSessionCacheId(id)) == null)
            {

                bool aa, bb;
                SessionPackage sessionp = new SessionPackage();
                sessionp.SessionId = id;
                sessionp.ApplicationName = ApplicationName;
                sessionp.Created = DateTime.Now;
                sessionp.Expires = DateTime.Now.AddMinutes(Convert.ToDouble(timeout));
                sessionp.LockDate = DateTime.Now;
                sessionp.LockId = Guid.NewGuid().ToString();
                sessionp.Timeout = timeout;
                sessionp.Locked = false;
                sessionp.SessionItems = MountSessionItemPack(context, id, out aa, out bb);
                sessionp.StaticObjects = MountStaticObjectPack(context, id);
                sessionp.Flags = 1;

                SessionCacheRemoveAction scra = new SessionCacheRemoveAction(m_endcallback, id, new SessionStateStoreData(sessionp.SessionItems, sessionp.StaticObjects, sessionp.Timeout));

                context.Cache.Insert(GetSessionCacheId(id),
                    sessionp,
                    null,
                    DateTime.MaxValue,
                    TimeSpan.FromMinutes(Convert.ToDouble(sessionp.Timeout)),
                    System.Web.Caching.CacheItemPriority.Low,
                    new System.Web.Caching.CacheItemRemovedCallback(scra.DoCacheRemoveEvent));

            }
        }


        //
        // SessionStateProviderBase.CreateNewStoreData
        //

        public override SessionStateStoreData CreateNewStoreData(
            HttpContext context,
            int timeout)
        {
            return new SessionStateStoreData(new SessionStateItemCollection(),
                SessionStateUtility.GetSessionStaticObjects(context),
                timeout);
        }



        //
        // SessionStateProviderBase.ResetItemTimeout
        //

        public override void ResetItemTimeout(HttpContext context,
            string id)
        {

            SessionPackage sessionp = (SessionPackage)context.Cache.Get(GetSessionCacheId(id));

            if (sessionp != null)
            {
                SessionCacheRemoveAction scra = new SessionCacheRemoveAction(m_endcallback, id, new SessionStateStoreData(sessionp.SessionItems, sessionp.StaticObjects, sessionp.Timeout));

                context.Cache.Insert(GetSessionCacheId(id),
                    sessionp,
                    null,
                    DateTime.MaxValue,
                    TimeSpan.FromMinutes(Convert.ToDouble(sessionp.Timeout)),
                    System.Web.Caching.CacheItemPriority.Low,
                    new System.Web.Caching.CacheItemRemovedCallback(scra.DoCacheRemoveEvent));
            }


        }


        //
        // SessionStateProviderBase.InitializeRequest
        //

        public override void InitializeRequest(HttpContext context)
        {
        }


        //
        // SessionStateProviderBase.EndRequest
        //

        public override void EndRequest(HttpContext context)
        {
        }




        private void ThreadedSave(object threadobj)
        {
            SaveSessionPack(((SessionPackage)threadobj).SessionItems, ((SessionPackage)threadobj).SessionId, ((SessionPackage)threadobj).StaticObjects);
        }


        private void ClearFilesAction(object stateobj)
        {
            string[] requestdata = stateobj.ToString().Split("\n".ToCharArray());

            System.IO.DirectoryInfo dirinfo = new System.IO.DirectoryInfo(requestdata[0]);
            System.IO.FileInfo[] finfos = dirinfo.GetFiles();

            foreach (System.IO.FileInfo finfo in finfos)
            {

                if (finfo.Exists)
                {
                    try
                    {
                        if (DateTime.Now.Subtract(finfo.LastAccessTime) >= TimeSpan.Parse(requestdata[1]))
                            finfo.Delete();
                    }
                    catch (System.IO.IOException)
                    {
                        //continue;
                    }
                    catch (System.UnauthorizedAccessException)
                    {
                        //continue;
                    }
                }
            }

            Array.Clear(finfos, 0, finfos.Length);
            finfos = null;

            GC.Collect();

        }

    }

    public class SessionPackage
    {
        public string SessionId = string.Empty;
        public string ApplicationName = string.Empty;
        public DateTime Created = DateTime.Now;
        public DateTime Expires = DateTime.Now.AddMinutes(20);
        public DateTime LockDate = DateTime.Now;
        public string LockId = Guid.NewGuid().ToString();
        public int Timeout = 20;
        public bool Locked = false;
        public SessionStateItemCollection SessionItems = new SessionStateItemCollection();
        public HttpStaticObjectsCollection StaticObjects = new HttpStaticObjectsCollection();
        public int Flags = 0;

        public SessionPackage(string id, string appname, SessionStateItemCollection items)
        {
            this.SessionId = id;
            this.ApplicationName = appname;
            this.SessionItems = items;
        }

        public SessionPackage(string id, string appname)
        {
            this.SessionId = id;
            this.ApplicationName = appname;
        }

        public SessionPackage()
        {
            this.SessionId = string.Empty;
        }

    }

    public class SessionCacheRemoveAction
    {
        private SessionStateItemExpireCallback sessionendhandler;
        private string sid = string.Empty;
        private SessionStateStoreData itemdata = null;

        public SessionCacheRemoveAction(SessionStateItemExpireCallback onsessionend, string id, SessionStateStoreData item)
        {
            this.sessionendhandler = onsessionend;
            this.sid = id;
            this.itemdata = item;
        }

        public void DoCacheRemoveEvent(String key, Object value, System.Web.Caching.CacheItemRemovedReason reason)
        {
            if (sessionendhandler != null)
                sessionendhandler(sid, itemdata);
        }
    }
}